function Xout = newton_manifold_r3_so3(fx,fX,RT,lineSinit,maxit)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Xout = gaussnewton_manifold_r3_so3(,fx,fX,RT_init,lineSinit,maxit)
%
% Is a function that refines the motion estimate via Newton
% Optimization on the Manifold. The parametrization is done over R3xSO3.
%
% Input:
%   
%   fx: A Matrix containing the 2D points on the image in the format 
%   (x1,y1,x2,y2,....). It is a vector of size 1x2N where N is the number 
%   of points. The coordinates are assumed to be normalized with the
%   intrinsic parameters of the camera.
%
%   fX: A Matrix containing the corresponsind 3D points to fx. It is of
%   dimenstion 1x3N where N is the number of points. It has the same format
%   of fx, i.e. (X1,Y1,Z1,X2,Y2,Z2,....)
%
%   RT_init: Initial estimate of the motion. It is a matrix of dimension 3x4 
%   of the form [R T] where R is of size 3x3 and T is of size 3x1
%
%   lineSinit: Method used to modify the Hessian. 0 for spectral
%   decomposition; otherwise, it uses the Choleksy factorization with added
%   multiple of the identity. This is the default method. 
%   Please refer to the book o Nocedal et al. for more details on such methods.
%
%   Note: The value of the modified Hessian value will be refined after the 
%   modification by finding the optimal step size while satisfying the 
%   Wolfe conditions. For that, we use the minFunc function provided by 
%   Mark Schmidt from ubc. The minFunc is a pre-requisite for the 
%   optimization here and need therefore to be found by the Matlab path 
%   before running this code. minFunc can be downloaded from (thanks to 
%   Mark Schmidt for making the code public): 
%   http://www.di.ens.fr/~mschmidt/Software/code.zip
%
%   maxit: Maximum Number of Iterations, default: 20.
%
% Output:
%   Xout: Refined 3D Motion in the form of a 3x4 matrix,
%   i.e [R T] where R is of size 3x3 and T is of size 3x1
%
% Author: Michel Sarkis
%   For Enquiries & bug reporting, please refer to the author at the 
%   following email: misarkis@gmail.com
%   Please also consider referring to our paper when using this code in any
%   publication: M. Sarkis & K. Diepold, "Camera Pose Estimation via 
%   Projective Newton Optimization on the Manifold", IEEE Trans. 
%   Image Processing, (Accepted for Publication)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%Check Input
if nargin < 2
    error('Algorithm requires at least the normalized points in the 2D images and their corresponding reconstructed 3D points.');
end

fX = fX(:);
if mod(length(fX), 3)~=0
    error 'Size of vector fX is of wrong dimension.'
end

num3DPoints = length(fX)/3;
if length(fx)/num3DPoints ~=2
    error 'Size of vector of image points fx is of wrong dimension.'
end

if nargin == 2    
    rt = rand(3,1);
    tt = rand(3,1);
    RT = [expm(skew(rt)) tt];
    lineSinit = 1;
    maxit = 20;    
    else if nargin == 3            
            if (size(RT,1)~=3 && size(RT,2)~=4)
                error 'Size of initial motion vector is of wrong dimension.'
            end          
            lineSinit = 1;
            maxit = 20;            
        else if nargin ==4                
                if (size(RT,1)~=3 && size(RT,2)~=4)
                    error 'Size of initial motion vector is of wrong dimension.'
                end      
                maxit = 20;
            end            
        end
end

%Get Parameter vector from rotation & translation
RT = generateCameraRotTransInv(RT);
x0 = convertCam2Parameter(RT);

%%Option for MinFunc which is called for step search
options.method = 'newton';
options.Display = 'off';
options.LS = 3;
options.TolX = 1e-12;
options.TolFun = 1e-12;

S1 = [eye(3,3) zeros(3,3)];
S2 = [zeros(3,3) eye(3,3)];
y1 = reshape(fX,3,[]);
T = x0(4:6);
R_tot = expm(skew(x0(1:3)));
    
    ex = [1:2:2*num3DPoints;2:2:2*num3DPoints];
    ex = ex(:);
    
    x1 = reshape(fx(ex),2,[]);
    
for a=1:maxit
    grad = zeros(6,1);
    hess1 = zeros(6,6);
    hess2 = zeros(6,6);
        
    for b=1:size(y1,2)
        if ~isnan(x1(1,b))
            x = [x1(:,b);1];
            z = R_tot*y1(:,b)+T;
            dd_dz = x'*z/(norm(x)*norm(z)^3).*z - x./(norm(x)*norm(z));
            d2d_dz2 = (z*x' + x*z' + x'*z.*eye(3,3))./(norm(x)*norm(z)^3) -...
                3*x'*z/(norm(x)*norm(z)^5).*z*z';
            grad = grad + (S1'*skew(R_tot*y1(:,b)) + S2')*dd_dz;
                
            hess2 = hess2 + (S2'-S1'*skew(R_tot*y1(:,b))')*d2d_dz2'...
                *(S2-skew(R_tot*y1(:,b))*S1);
            hess1 = hess1 + S1'*skew(R_tot*y1(:,b))*skew(dd_dz)*S1;     
            
        end
    end
         
    hess = hess1 + hess2;
        
    if(lineSinit==0)            
        % Modified Spectral Decomposition for the computation of the 
        %initial value of the Hessian 
        [V,D] = eig((hess+hess')/2);     
        D = diag(D);
        D = max(abs(D),max(max(abs(D)),1)*1e-12);
        % Calculation of the initial Hessian value, i.e equation 24 in
        % our paper
        d = V*((V'*grad)./D);
    else
        % Cholesky Factorization with added multiple of identities
        % for the computation of the initial value of the Hessian 
        beta = 1e-3;
        D = diag((hess+hess')/2);
        aii = min(D);     
        tk = 0;
        if(aii<0)            
            tk = -aii+beta;
        end
        %t0
        p = 1;
        while(p~=0)
            [R,p] = chol((hess+hess')/2+tk*eye(size(hess)));
                if p>0                                       
                    tk = max(2*tk,beta);
                end
        end
        % Calculation of the initial Hessian value, i.e equation 24 in
        % our paper
        d = (R'*R)\grad;            
    end         
        
    % Optimal Stepsize calculation that satisfies Wolfeconditions in order 
    % to update the motion values. This is done by calling the function 
    % minFunc as mentioned previously.
    [t fn] = minFunc(@funcval, 0, options, d, R_tot, x1, y1, T);        
    R1 = expm(-t*skew(S1*d));
    T1 = -t*S2*d;        

    %update Variables
    T = T + T1;
    R_tot = R1*R_tot;
        
end

Xout = [R_tot T];


% Auxiliary function to calculate Hessian & gradient needed for the optimal
% step size calculation
function [res, d1, d2] = funcval(t, grad, R_tot, x1, y1, T)
res = 0; d1 = 0; d2 = 0;
R = expm(-skew(t*grad(1:3)));
T1 = T - t*grad(4:6);
S1 = [eye(3,3) zeros(3,3)];
S2 = [zeros(3,3) eye(3,3)];
for b=1:size(y1,2)
    if ~isnan(x1(1,b))
        x = [x1(:,b);1];
        z = R*R_tot*y1(:,b)+T1;
        yi = R_tot*y1(:,b);
        res = res + cosdist(x, z);
        if nargout > 1
            dd_dz = (x'*z/(norm(x)*norm(z)^3).*z - x./(norm(x)*norm(z)));
            d1 = d1 + dd_dz'*(R*skew(yi)*S1 - S2)*grad;

            if nargout > 2

                d2d_dz2 = z*(x./(norm(x)*norm(z)^3) - 3*x'*z/(norm(x)*norm(z)^5).*z)' ...
                    + x'*z/(norm(x)*norm(z)^3)*eye(3,3) + x*z'./(norm(x)*norm(z)^3);
                d2 = d2 + grad'*(R*skew(yi)*S1-S2)'*d2d_dz2'*(R*skew(yi)*S1-S2)*grad ...
                    + grad'*S1'*skew(dd_dz)*R*skew(yi)*S1*grad;
            end
        end
    end
end

% Auxiliary function to calculate the Cosine distance
function retval = cosdist(x,y)
retval = 1-x(:)'*y(:)/(norm(x)*norm(y));